/* ClientSocket.c - NeonServ v5.6
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#include "main.h"
#include "ClientSocket.h"
#include "IRCParser.h"
#include "UserNode.h"
#include "IRCQueue.h"
#include "WHOHandler.h"
#include "HandleInfoHandler.h"
#include "ConfigParser.h"
#include "version.h"
#include "IOHandler.h"
#include "IRCEvents.h"
#include "log.h"

struct socket_list {
    struct ClientSocket *data;
    unsigned count;
};

#ifdef HAVE_THREADS
static pthread_mutex_t synchronized;
static pthread_mutex_t synchronized_recv;

struct ParseOrder {
    unsigned int tid;
    struct ParseOrder *next;
};
struct ParseOrder *parse_order = NULL;
#endif

//the magic list :P
static struct socket_list *sockets = NULL;

static IOHANDLER_CALLBACK(socket_callback);

void init_sockets() {
    THREAD_MUTEX_INIT(synchronized);
    THREAD_MUTEX_INIT(synchronized_recv);
    
    sockets = malloc(sizeof(*sockets));
    if (!sockets)
    {
        printf_log("main", LOG_ERROR, "%s:%d malloc() failed", __FILE__, __LINE__);
        return;
    }
    sockets->data = NULL;
    sockets->count = 0;
}

struct ClientSocket* create_socket(char *host, int port, char *bindto, char *pass, char *nick, char *ident, char *realname) {
    if(!sockets)
        init_sockets();
    struct ClientSocket *client = malloc(sizeof(*client));
    if (!client) {
        return NULL;
    }
    client->iofd = NULL;
    client->host = strdup(host);
    client->port = port;
    client->bind = (bindto ? strdup(bindto) : NULL);
    client->pass = (pass == NULL ? NULL : strdup(pass));
    client->nick = strdup(nick);
    client->ident = strdup(ident);
    client->realname = strdup(realname);
    client->user = NULL;
    client->network_name = NULL;
    client->flags = 0;
    client->traffic_in = 0;
    client->traffic_out = 0;
    client->connection_time = 0;
	client->botid = 0;
    client->clientid = 0;
    client->queue = NULL;
    client->whoqueue_first = NULL;
    client->whoqueue_last = NULL;
    client->handleinfo_first = NULL;
    client->handleinfo_last = NULL;
    SYNCHRONIZE(synchronized);
    client->next = sockets->data;
    sockets->data = client;
    DESYNCHRONIZE(synchronized);
    return client;
}

void connect_socket(struct ClientSocket *client) {
    client->iofd = iohandler_connect(client->host, client->port, ((client->flags & SOCKET_FLAG_SSL) ? 1 : 0), client->bind, socket_callback);
    client->iofd->data = client;
    client->flags |= SOCKET_FLAG_RECONNECT;
}

static int _close_socket(struct ClientSocket *client) {
    if(client == NULL) return 0;
    if((client->flags & SOCKET_FLAG_CONNECTED)) {
        iohandler_printf(client->iofd, "QUIT :[NeonServ %s.%d] disconnect requested.\n", NEONSERV_VERSION, patchlevel);
        
        bot_disconnect(client);
        
        iohandler_close(client->iofd);
        client->iofd = NULL;
    }
    client->flags &= ~(SOCKET_FLAG_READY | SOCKET_FLAG_CONNECTED);
    if(client->queue)
        queue_destroy(client);
    if(client->whoqueue_first)
        clear_whoqueue(client);
    if(client->handleinfo_first)
        clear_handleinfoqueue(client);
    return 1;
}

int close_socket(struct ClientSocket *client) { //external call (don't reconnect)
    if(client == NULL) return 0;
    client->flags &= ~SOCKET_FLAG_RECONNECT;
    return _close_socket(client);
}

int destroy_socket(struct ClientSocket *client) {
    if(client == NULL) return 0;
    close_socket(client);
    SYNCHRONIZE(synchronized);
    struct ClientSocket *sock, *last_sock = NULL;
    for (sock = sockets->data; sock; sock = sock->next) {
        if(sock == client) {
            if(last_sock)
                last_sock->next = sock->next;
            else
                sockets->data = sock->next;
            sockets->count--;
            break;
        } else
            last_sock = sock;
    }
    event_freeclient(client);
    free(client->host);
    if(client->bind)
        free(client->bind);
    if(client->pass)
        free(client->pass);
    if(client->network_name)
        free(client->network_name);
    if(client->iofd) //reconnect timer?
        iohandler_close(client->iofd);
    free(client);
    DESYNCHRONIZE(synchronized);
    return 1;
}

int write_socket_force(struct ClientSocket *client, char* msg, int len) {
    if(!(client && (client->flags & SOCKET_FLAG_CONNECTED))) return 0;
    SYNCHRONIZE(synchronized);
    #ifdef HAVE_THREADS
    printf_log("main", LOG_IRCRAW, "[%d send %d] %s", getCurrentThreadID(), len, msg);
    #else
    printf_log("main", LOG_IRCRAW, "[send %d] %s", len, msg);
    #endif
	iohandler_send(client->iofd, msg, len);
    client->traffic_out += len;
    DESYNCHRONIZE(synchronized);
    return 1;
}

int write_socket(struct ClientSocket *client, char* msg, int len) {
    if(!(client && (client->flags & SOCKET_FLAG_CONNECTED))) return 0;
    if(client->flags & SOCKET_FLAG_USE_QUEUE)
        return queue_add(client, msg, len);
    else
        return write_socket_force(client, msg, len);
}

#if HAVE_THREADS
static void clientsocket_start_of_recv(unsigned int tid) {
    SYNCHRONIZE(whohandler_sync);
    struct ParseOrder *entry, *last;
    for(last = parse_order; last; last = last->next) {
        if(last->next == NULL)
            break;
    }
    entry = malloc(sizeof(*entry));
    entry->tid = tid;
    entry->next = NULL;
    if(last)
        last->next = entry;
    else
        parse_order = entry;
    DESYNCHRONIZE(whohandler_sync);
}

static void clientsocket_end_of_recv(unsigned int tid) {
    SYNCHRONIZE(whohandler_sync);
    struct ParseOrder *entry, *last = NULL;
    for(entry = parse_order; entry; entry = entry->next) {
        if(entry->tid == tid) {
            if(last)
                last->next = entry->next;
            else
                parse_order = entry->next;
            free(entry);
            break;
        } else
            last = entry;
    }
    DESYNCHRONIZE(whohandler_sync);
}

int clientsocket_parseorder_top(unsigned int tid) {
    if(parse_order && parse_order->tid == tid)
        return 1;
    else
        return 0;
}
#endif

static IOHANDLER_CALLBACK(socket_callback) {
    struct ClientSocket *client = event->iofd->data;
    #ifdef HAVE_THREADS
    unsigned int tid;
    #endif
    if(process_state.running == 0)
        return; //just ignore the event (shutdown sequence)
    switch(event->type) {
    case IOEVENT_CONNECTED:
        client->flags |= SOCKET_FLAG_CONNECTED;
        if(client->pass && strcmp(client->pass, ""))
            putsock(client, "PASS :%s", client->pass);
        putsock(client, "USER %s 0 0 :%s", client->ident, client->realname);
        putsock(client, "NICK %s", client->nick);
        break;
    case IOEVENT_NOTCONNECTED:
    case IOEVENT_CLOSED:
        _close_socket(client);
        if(client->flags & SOCKET_FLAG_RECONNECT) {
            struct timeval timeout;
            gettimeofday(&timeout, NULL);
            timeout.tv_sec += SOCKET_RECONNECT_TIME;
            client->iofd = iohandler_timer(timeout, socket_callback);
            client->iofd->data = client;
        }
        break;
    case IOEVENT_TIMEOUT: //reconnect timer
        connect_socket(client);
        break;
    case IOEVENT_RECV:
        #ifdef HAVE_THREADS
        tid = (unsigned int) pthread_self_tid();
        clientsocket_start_of_recv(tid);
        #endif
        client->traffic_in += strlen(event->data.recv_str);
        parse_line(client, event->data.recv_str);
        #ifdef HAVE_THREADS
        clientsocket_end_of_recv(tid);
        #endif
        break;
    default:
        break;
    }
}

void putsock(struct ClientSocket *client, const char *text, ...) {
    va_list arg_list;
    char sendBuf[MAXLEN];
    int pos;
    if (!(client && (client->flags & SOCKET_FLAG_CONNECTED))) return;
    sendBuf[0] = '\0';
    va_start(arg_list, text);
    pos = vsnprintf(sendBuf, MAXLEN - 2, text, arg_list);
    va_end(arg_list);
    if (pos < 0 || pos > (MAXLEN - 2)) pos = MAXLEN - 2;
    sendBuf[pos] = '\n';
    sendBuf[pos+1] = '\0';
    write_socket(client, sendBuf, pos+1);
}

struct ClientSocket* getBots(int flags, struct ClientSocket* last_bot) {
    struct ClientSocket *sock = (last_bot ? last_bot->next : sockets->data);
    if(sock == NULL) return NULL;
    for (; sock; sock = sock->next) {
        if(!flags || (sock->flags & flags) == flags)
            return sock;
    }
    return NULL;
}

void free_sockets(int close_only) {
    if(!sockets) return;
    struct ClientSocket *client, *next;
    for (client = sockets->data; client; client = next) {
        next = client->next;
        if(close_only) {
            if((client->flags & SOCKET_FLAG_CONNECTED))
                iohandler_printf(client->iofd, "QUIT :[NeonServ %s.%d] shutdown requested.\n", NEONSERV_VERSION, patchlevel);
        } else
            destroy_socket(client);
    }
    if(!close_only) {
        free(sockets);
        sockets = NULL;
    }
}
